@@ -73,10 +73,10 @@ def Plot(
             if as_curve is None:
                 sites = list(details.sites)
             else:
-                max_arc_length = as_curve[0].x.item(-1)
+                max_arc_length = as_curve.x.item(-1)
                 n_samples = int(round(_SUBSAMPLING_FACTOR * details.measures.length))
                 arc_lengths = nmpy.linspace(0.0, max_arc_length, num=max(n_samples, 2))
-                sites = [as_curve[idx_](arc_lengths) for idx_ in range(space_dim)]
+                sites = [_row for _row in as_curve(arc_lengths)]
         elif mode is plot_mode_e.SKL_Pixel:
             sites = list(details.sites)
         else:
@@ -97,13 +97,13 @@ def Plot(
         )
 
         if direction_style.show:
-            origin_direction = details.OriginDirection()
-            if origin_direction is not None:
+            initial_direction = details.InitialDirection()
+            if initial_direction is not None:
                 dir_sites = tuple(
                     nmpy.hstack((sites[idx_][0], sites[idx_][-1]))
                     for idx_ in range(space_dim)
                 )
-                directions = list(zip(origin_direction, details.FinalDirection()))
+                directions = list(zip(initial_direction, details.FinalDirection()))
                 directions[0], directions[1] = (
                     directions[1],
                     TransformedYForVector(directions[0]),
