@@ -48,7 +48,7 @@ from skl_graph.type.node import branch_node_t, end_node_t, node_t
 array_t = nmpy.ndarray
 
 
-_MAX_COORDINATE_INTERPOLATION_ERROR = 0.8
+_MAX_SQ_INTERPOLATION_ERROR = 0.8**2
 
 
 class measures_t(named_tuple_t):
@@ -119,61 +119,67 @@ class edge_t(raw_edge_t):
 
         return transient.AsEdge()
 
-    def AsCurve(self) -> tuple | None:
+    def AsCurve(self) -> spnt.PchipInterpolator | None:
         """"""
         cache_entry = self.AsCurve.__name__
 
         if cache_entry not in self._cache:
             if self.n_sites > 1:
-                as_list: list[float] = list(self.measures.segment_lengths)
-                arc_lengths = nmpy.cumsum([0.0] + as_list)
-                self._cache[cache_entry] = tuple(
-                    _PchipInterpolator(arc_lengths, self.sites[idx_])
-                    for idx_ in range(self.dim)
+                arc_lengths = nmpy.cumsum([0.0] + list(self.measures.segment_lengths))
+                self._cache[cache_entry] = _PchipInterpolator(
+                    arc_lengths, nmpy.array(self.sites)
                 )
             else:
                 self._cache[cache_entry] = None
 
         return self._cache[cache_entry]
 
-    def OriginDirection(self) -> array_t | None:
+    def InitialDirection(self) -> array_t | None:
         """"""
-        return self._Direction(self.OriginDirection)
+        return self._Direction(self.InitialDirection)
 
     def FinalDirection(self) -> array_t | None:
         """"""
         return self._Direction(self.FinalDirection)
 
-    def _Direction(self, where: Callable, /) -> array_t:
+    def _Direction(self, caller: Callable, /) -> array_t:
         """"""
-        where = where.__name__
-        if where not in self._cache:
-            self._SetEndPointDirections()
+        cache_entry = caller.__name__
+        if cache_entry not in self._cache:
+            cache_entry_i = self.InitialDirection.__name__
+            cache_entry_f = self.FinalDirection.__name__
 
-        return self._cache[where]
+            as_curve = self.AsCurve()
+            if as_curve is None:
+                self._cache[cache_entry_i] = None
+                self._cache[cache_entry_f] = None
+            else:
+                dim_range = tuple(range(self.sites.__len__()))
+                first_inner = tuple(
+                    self.sites[_idx][1] - self.sites[_idx][0] for _idx in dim_range
+                )
+                last_inner = tuple(
+                    self.sites[_idx][-2] - self.sites[_idx][-1] for _idx in dim_range
+                )
 
-    def _SetEndPointDirections(self) -> None:
-        """"""
-        cache_entry_o = self.OriginDirection.__name__
-        cache_entry_f = self.FinalDirection.__name__
+                max_arc_length = as_curve.x.item(-1)
+                derivatives = as_curve((0, max_arc_length), 1)
+                initial = derivatives[:, 0]
+                final = derivatives[:, 1]
+                flips = [1.0, 1.0]
+                for p_idx, product in enumerate(
+                    (nmpy.dot(initial, first_inner), nmpy.dot(final, last_inner))
+                ):
+                    if product < 0.0:
+                        flips[p_idx] = -1.0
+                self._cache[cache_entry_i] = initial / (
+                    flips[0] * nmpy.linalg.norm(initial)
+                )
+                self._cache[cache_entry_f] = final / (
+                    flips[1] * nmpy.linalg.norm(final)
+                )
 
-        as_curve = self.AsCurve()
-        if as_curve is None:
-            self._cache[cache_entry_o] = None
-            self._cache[cache_entry_f] = None
-        else:
-            max_arc_length = as_curve[0].x.item(-1)
-            o_dir, f_dir = [], []
-            for d_idx in range(self.dim):
-                directions = as_curve[d_idx]((0, max_arc_length), 1)
-                o_dir.append(directions[0])
-                f_dir.append(directions[1])
-            self._cache[cache_entry_o] = nmpy.array(o_dir, dtype=nmpy.float64) / (
-                -nmpy.linalg.norm(o_dir)
-            )
-            self._cache[cache_entry_f] = nmpy.array(
-                f_dir, dtype=nmpy.float64
-            ) / nmpy.linalg.norm(f_dir)
+        return self._cache[cache_entry]
 
     def Curvatures(self) -> tuple[float, ...]:
         """"""
@@ -186,23 +192,24 @@ class edge_t(raw_edge_t):
 
         if cache_entry not in self._cache:
             as_curve = self.AsCurve()
-            arc_lengths = as_curve[0].x
-            # derivatives_?: first index=0, 1 for derivative orders 1, 2; second index=arc-length parameter
-            derivatives_x, derivatives_y = (
-                tuple(as_curve[_idx](arc_lengths, _rdr) for _rdr in (1, 2))
-                for _idx in range(self.dim)
-            )
+            if as_curve is None:
+                self._cache[cache_entry] = None
+            else:
+                arc_lengths = as_curve.x
+                derivatives_fst, derivatives_scd = tuple(
+                    as_curve(arc_lengths, _rdr) for _rdr in (1, 2)
+                )
 
-            normalization = (derivatives_x[0] ** 2 + derivatives_y[0] ** 2) ** 1.5
-            where_safe = normalization != 0.0
-            curvatures = nmpy.empty_like(arc_lengths)
-            curvatures[where_safe] = (
-                derivatives_x[0][where_safe] * derivatives_y[1][where_safe]
-                - derivatives_y[0][where_safe] * derivatives_x[1][where_safe]
-            ) / normalization[where_safe]
-            curvatures[nmpy.logical_not(where_safe)] = 0.0
+                normalization = nmpy.sum(derivatives_fst**2, axis=0) ** 1.5
+                where_safe = normalization != 0.0
+                curvatures = nmpy.empty_like(arc_lengths)
+                curvatures[where_safe] = (
+                    derivatives_fst[0][where_safe] * derivatives_scd[1][where_safe]
+                    - derivatives_fst[1][where_safe] * derivatives_scd[0][where_safe]
+                ) / normalization[where_safe]
+                curvatures[nmpy.logical_not(where_safe)] = 0.0
 
-            self._cache[cache_entry] = tuple(curvatures)
+                self._cache[cache_entry] = tuple(curvatures)
 
         return self._cache[cache_entry]
 
@@ -217,109 +224,55 @@ class edge_t(raw_edge_t):
 
         if cache_entry not in self._cache:
             as_curve = self.AsCurve()
-            arc_lengths = as_curve[0].x
-            # derivatives_?: first index=0, 1, 2 for derivative orders 1, 2, 3; second index=arc-length parameter
-            derivatives_x, derivatives_y, derivatives_z = (
-                tuple(as_curve[_idx](arc_lengths, _rdr) for _rdr in (1, 2, 3))
-                for _idx in range(self.dim)
-            )
-            # key=space dimension index among 1, 2, 3 (i.e., X, Y, Z)
-            # value=list of first+second+third derivatives for each curve parameter
-            # derivatives: dict[int, list[tuple[float, float, float]]] = {}
-            # for d_idx in range(self.dim):
-            #     per_dim = []
-            #     for parameter in arc_lengths:
-            #         derivative_1 = as_curve[d_idx](parameter, 1).item()
-            #         derivative_2 = as_curve[d_idx](parameter, 2).item()
-            #         derivative_3 = as_curve[d_idx](parameter, 3).item()
-            #         per_dim.append((derivative_1, derivative_2, derivative_3))
-            #     derivatives[d_idx + 1] = per_dim
-
-            a = (
-                derivatives_z[1] * derivatives_y[0]
-                - derivatives_y[1] * derivatives_z[0]
-            )
-            b = (
-                derivatives_x[1] * derivatives_z[0]
-                - derivatives_z[1] * derivatives_x[0]
-            )
-            c = (
-                derivatives_y[1] * derivatives_x[0]
-                - derivatives_x[1] * derivatives_y[0]
-            )
-            abc_sq_length = a**2 + b**2 + c**2
-
-            normalization = (
-                derivatives_x[0] ** 2 + derivatives_y[0] ** 2 + derivatives_z[0] ** 2
-            ) ** 1.5
-            where_safe = normalization != 0.0
-            curvatures = nmpy.empty_like(arc_lengths)
-            curvatures[where_safe] = (
-                nmpy.sqrt(abc_sq_length[where_safe]) / normalization[where_safe]
-            )
-            curvatures[nmpy.logical_not(where_safe)] = 0.0
-
-            where_safe = abc_sq_length != 0.0
-            torsions = nmpy.empty_like(arc_lengths)
-            # Why abs? From: https://mathworld.wolfram.com/Torsion.html
-            #     The torsion is positive for a right-handed curve, [...]
-            # From : https://mathcurve.com/courbes3d.gb/dextre/dextre.shtml
-            #     A right-handed curve is a curve with positive torsion
-            # (!)
-            # Anyway, this explains why abs.
-            torsions[where_safe] = nmpy.abs(
-                (
-                    derivatives_x[2][where_safe] * a[where_safe]
-                    + derivatives_y[2][where_safe] * b[where_safe]
-                    + derivatives_z[2][where_safe] * c[where_safe]
+            if as_curve is None:
+                self._cache[cache_entry] = None
+            else:
+                arc_lengths = as_curve.x
+                derivatives_fst, derivatives_scd, derivatives_trd = tuple(
+                    as_curve(arc_lengths, _rdr) for _rdr in (1, 2, 3)
                 )
-                / abc_sq_length[where_safe]
-            )
-            torsions[nmpy.logical_not(where_safe)] = 0.0
 
-            # curvatures = []
-            # torsions = []
-            # for p_idx in range(arc_lengths.__len__()):
-            #     a = (
-            #         derivatives_z[1][p_idx] * derivatives_y[0][p_idx]
-            #         - derivatives_y[1][p_idx] * derivatives_z[0][p_idx]
-            #     )
-            #     b = (
-            #         derivatives_x[1][p_idx] * derivatives_z[0][p_idx]
-            #         - derivatives_z[1][p_idx] * derivatives_x[0][p_idx]
-            #     )
-            #     c = (
-            #         derivatives_y[1][p_idx] * derivatives_x[0][p_idx]
-            #         - derivatives_x[1][p_idx] * derivatives_y[0][p_idx]
-            #     )
-            #     abc_sq_length = a**2 + b**2 + c**2
-            #
-            #     normalization = (
-            #         derivatives_x[0][p_idx] ** 2
-            #         + derivatives_y[0][p_idx] ** 2
-            #         + derivatives_z[0][p_idx] ** 2
-            #     ) ** 1.5
-            #     if normalization != 0.0:
-            #         curvature = (nmpy.sqrt(abc_sq_length) / normalization).item()
-            #     else:
-            #         curvature = 0.0
-            #
-            #     if abc_sq_length != 0.0:
-            #         torsion = nmpy.abs(
-            #             (
-            #                 derivatives_x[2][p_idx] * a
-            #                 + derivatives_y[2][p_idx] * b
-            #                 + derivatives_z[2][p_idx] * c
-            #             )
-            #             / abc_sq_length
-            #         ).item()
-            #     else:
-            #         torsion = 0.0
-            #
-            #     curvatures.append(curvature)
-            #     torsions.append(torsion)
+                a = (
+                    derivatives_scd[2] * derivatives_fst[1]
+                    - derivatives_scd[1] * derivatives_fst[2]
+                )
+                b = (
+                    derivatives_scd[0] * derivatives_fst[2]
+                    - derivatives_scd[2] * derivatives_fst[0]
+                )
+                c = (
+                    derivatives_scd[1] * derivatives_fst[0]
+                    - derivatives_scd[0] * derivatives_fst[1]
+                )
+                abc_sq_length = a**2 + b**2 + c**2
 
-            self._cache[cache_entry] = (tuple(curvatures), tuple(torsions))
+                normalization = nmpy.sum(derivatives_fst**2, axis=0) ** 1.5
+                where_safe = normalization != 0.0
+                curvatures = nmpy.empty_like(arc_lengths)
+                curvatures[where_safe] = (
+                    nmpy.sqrt(abc_sq_length[where_safe]) / normalization[where_safe]
+                )
+                curvatures[nmpy.logical_not(where_safe)] = 0.0
+
+                where_safe = abc_sq_length != 0.0
+                torsions = nmpy.empty_like(arc_lengths)
+                # Why abs? From: https://mathworld.wolfram.com/Torsion.html
+                #     The torsion is positive for a right-handed curve, [...]
+                # From : https://mathcurve.com/courbes3d.gb/dextre/dextre.shtml
+                #     A right-handed curve is a curve with positive torsion
+                # (!)
+                # Anyway, this explains why abs.
+                torsions[where_safe] = nmpy.abs(
+                    (
+                        derivatives_trd[0][where_safe] * a[where_safe]
+                        + derivatives_trd[1][where_safe] * b[where_safe]
+                        + derivatives_trd[2][where_safe] * c[where_safe]
+                    )
+                    / abc_sq_length[where_safe]
+                )
+                torsions[nmpy.logical_not(where_safe)] = 0.0
+
+                self._cache[cache_entry] = (tuple(curvatures), tuple(torsions))
 
         return self._cache[cache_entry]
 
@@ -691,20 +644,18 @@ def _LMapNeighborhood(lmap: array_t, site: tuple[int, ...]) -> tuple[array_t, ar
 
 
 def _PchipInterpolator(
-    arc_lengths: array_t, coordinates: array_t, /
+    arc_lengths: array_t, sites: array_t, /
 ) -> spnt.PchipInterpolator:
     """"""
     for n_samples in range(2, arc_lengths.size + 1):
-        samples = nmpy.linspace(0, arc_lengths.size - 1, num=n_samples)
-        indices = nmpy.unique(nmpy.around(samples)).astype(nmpy.uint64)
+        arc_length_samples = nmpy.linspace(0, arc_lengths.size - 1, num=n_samples)
+        indices = nmpy.unique(nmpy.around(arc_length_samples)).astype(nmpy.uint64)
         output = spnt.PchipInterpolator(
-            arc_lengths[indices], coordinates[indices], extrapolate=False
+            arc_lengths[indices], sites[:, indices], axis=1, extrapolate=False
         )
 
-        error = max(abs(output(arc_lengths) - coordinates))
-        if error <= _MAX_COORDINATE_INTERPOLATION_ERROR:
+        error = max(nmpy.sum((output(arc_lengths) - sites) ** 2, axis=0))
+        if error <= _MAX_SQ_INTERPOLATION_ERROR:
             return output
 
-    raise RuntimeError(
-        f"{arc_lengths} -> {coordinates}: Unable to interpolate with sufficient precision."
-    )
+    raise RuntimeError("Should never happen. Please contact developer.")
